structure Unification = 
struct
    datatype term =
      Var of string
    | Pred of string * term list

    datatype rule = Rule of term * term list

    datatype constraint = Eq of term * term

    datatype unifier = VarEq of string * term

    (* printing functions *)
    fun pp_list_nobrack (pp : 'a -> string) (xs : 'a list) : string =
    case xs of
        [] => ""
    | [x] => pp x
    | x::xs => (pp x) ^ ", " ^ (pp_list_nobrack pp xs)

    fun pp_list (pp : 'a -> string) (xs : 'a list) : string =
    "[" ^ (pp_list_nobrack pp xs) ^ "]"
    
    fun pp_term t = case t of
        Var s => "?" ^ s
    | Pred (f, ts) => f ^ "(" ^ (pp_list_nobrack (pp_term) ts) ^ ")"

    fun pp_rule (Rule (t, ts)) =
    (pp_term t) ^ " :- " ^ (pp_list (pp_term) ts) ^ "."

    fun pp_constraint (Eq(t1, t2)) =
    (pp_term t1) ^ " = " ^ (pp_term t2)

    fun pp_unifier (VarEq(s, t)) =
    s ^ " ::= " ^ (pp_term t)

    (* substitution functions *)
    fun subst_term (u as VarEq(a, t1) : unifier)
          (t2 : term)
        : term =
      case t2 of
        Var b => if a = b then t1 else t2
      | Pred (f, ts) => Pred (f, map (subst_term u) ts)

    fun subst_terms (u : unifier)
                    (ts : term list)
        : term list =
      List.map (subst_term u) ts

    fun subst_all (us : unifier list) (terms : term list) =
      List.foldl (fn (u, ts) => subst_terms u ts) terms us


    (* fresh rule creation *)
    val counter = ref 15316
    fun new_string () =
      (counter := !counter + 1; 
      "_var" ^ (Int.toString (!counter)))

    fun new_var () = Var (new_string ())

    fun unique (l : ''a list) : ''a list =
      case l of
        [] => []
      | x::xs => x::(unique (List.filter (fn y => y <> x) xs))

    fun all_vars (ts : term list) : term list =
      case ts of
        [] => []
      | (Var s)::ts' => (Var s)::(all_vars ts')
      | (Pred (f, args))::ts' => (all_vars args) @ (all_vars ts')

    fun vars ts = unique (all_vars ts)

    fun fresh_rule (Rule (t, ts)) = 
      let val vs = vars (t::ts)
          val new_vs = List.map (fn Var s => (VarEq(s, new_var ()))) vs
          val t'::ts' = (subst_all new_vs (t::ts))
      in Rule(t', ts')
      end

end